import math
import pathlib
import itertools
import collections

from lingpy.sequence.sound_classes import class2tokens
from lingpy.settings import rc
from lingpy.align.sca import get_consensus, Alignments
from lingpy.util import pb
from lingpy import log
from lingpy import basictypes as bt

import networkx as nx


def consensus_pattern(patterns, missing="Ø"):
    """
    Return consensus pattern of multiple patterns.

    :param patterns: list of patterns
    :param missing: the character used to represent missing values

    .. note:: This consensus method raises an error if the patterns contain incompatible
       columns (non-identical values apart from the missing data character in the same
       column).
    """
    out = []
    for i in range(len(patterns[0])):
        col = [line[i] for line in patterns]
        no_gaps = [x for x in col if x != missing]
        if len(set(no_gaps)) > 1:
            raise ValueError("Your patterns are incompatible")
        out += [no_gaps[0] if no_gaps else missing]
    return tuple(out)


def incompatible_columns(patterns, missing="Ø"):
    """
    Compute whether a pattern has incompatible columns.
    """
    columns = []
    for i in range(len(patterns[0])):
        col = [
            patterns[j][i] for j in range(len(patterns)) if patterns[j][i] != missing
        ]
        columns.append("*" if len(set(col)) > 1 else "")
    return columns


def score_patterns(patterns, missing="Ø", mode="coverage"):
    """
    Function gives a score for the overall number of reflexes.

    .. note:: This score tells simply to which degree a pattern is filled. It divides the
       number of cells not containing missing data by the number of cells in the
       matrix.
    """
    # return -1 if the patterns are not compatible
    for i in range(len(patterns[0])):
        if len(set([row[i] for row in patterns if row[i] != missing])) > 1:
            return -1
    if len(patterns) <= 1:
        return -1

    if mode not in ["ranked", "pairs", "squared", "coverage"]:
        raise ValueError("you must select an appropriate mode")

    # we rank the columns by sorting them first
    if mode == "ranked":
        cols = []
        for i in range(len(patterns[0])):
            cols += [sum([0 if row[i] == missing else 1 for row in patterns])]
        # sort the columns
        ranks, cols = list(range(1, len(cols) + 1))[::-1], sorted(cols, reverse=True)
        scores = []
        for rank, col in zip(ranks, cols):
            scores += [rank * col]
        return sum(scores) / sum(ranks) / len(patterns)

    if mode == "squared":
        psize = len(patterns[0])
        scores = [((psize - row.count(missing)) / psize) ** 2 for row in patterns]
        return sum(scores) / len(scores)

    if mode == "pairs":

        # count the number of pairs in the data
        pairs = 0
        covered = 0
        m, n = len(patterns[0]), len(patterns)
        for i in range(n):
            vals = m - patterns[i].count(missing)
            pairs += (vals ** 2 - vals) / 2
        for i in range(m):
            vals = n - [p[i] for p in patterns].count(missing)
            pairs += (vals ** 2 - vals) / 2
            if vals != 0:
                covered += 1
        return ((pairs / n) / covered) / m

    if mode == "coverage":
        cols = []
        for i in range(len(patterns[0])):
            col = [row[i] for row in patterns]
            cols += [len(patterns) - col.count(missing)]
        return (sum(cols) / len(patterns[0])) / len(patterns)  # * len(patterns[0]))


def compatible_columns(colA, colB, missing="Ø", gap="-"):
    """Check for column compatibility.

    Parameters
    ----------
    colA, colB = list
        Lists (sequence type) containing a given pattern.
    missing : str (default="Ø")
        A gap in the sense of "missing data", that is, a cognate set for which
        a value in a given language is absent.

    Returns
    -------
    matches, mismatches : tuple
        The score for matches gives zero if there is no conflict but also no
        match. For mismatches it is accordingly. So if you seek for
        compatibility, a mismatch greater 0 means the patterns are not
        compatible.
    """
    matches, mismatches = 0, 0
    for a, b in zip(colA, colB):
        if missing not in [a, b]:
            if a != b:
                mismatches += 1
            else:
                if a != gap:
                    matches += 1
    return matches, mismatches


def density(wordlist, ref="cogid"):
    """Compute the density of a wordlist.


    XXX TOTO: better move to util or elsewhere

    Notes
    -----
    We define the density of a wordlist by measuring how many words can be
    explained by the same cognate set.
    """
    scores = []
    for concept in wordlist.rows:
        idxs = wordlist.get_list(row=concept, flat=True)
        cogids = [wordlist[idx, ref] for idx in idxs]
        sums = [1 / cogids.count(cogid) for idx, cogid in zip(idxs, cogids)]
        scores.append(sum(sums) / len(sums))
    return 1 - sum(scores) / len(scores)


class CoPaR(Alignments):
    """Correspondence Pattern Recognition class

    Parameters
    ----------
    wordlist : ~lingpy.basic.wordlist.Wordlist
        A wordlist object which should have a column for segments and a column
        for cognate sets. Since the class inherits from LingPy's
        Alignments-class, the same kind of data should be submitted.
    ref : str (default="cogid")
        The column which stores the cognate sets.
    segments : str (default="tokens")
        The column which stores the segmented transcriptions.
    alignment : str (default="alignment")
        The column which stores the alignments (or will store the alignments if
        they have not yet been computed).
    """

    def __init__(
        self,
        wordlist,
        minrefs=3,
        ref="cogids",
        structure="structure",
        missing="Ø",
        gap="-",
        irregular="!?",
        **keywords
    ):
        Alignments.__init__(self, wordlist, ref=ref, **keywords)
        self.ref = ref
        self._structure = structure
        self.minrefs = minrefs
        self.missing = missing
        self.gap = gap
        self.irregular = irregular
        if structure not in self.columns:
            raise ValueError("no column {0} for structure was found".format(structure))

    def positions_from_prostrings(self, cogid, indices, alignment, structures):
        """
        Return positions matching from an alignment and user-defined prosodic strings
        """
        if self._mode == "fuzzy":
            strucs = []
            for idx, struc, alm in zip(indices, structures, alignment):
                pos_ = self[idx, self._ref].index(cogid)
                strucs += [class2tokens(struc.n[pos_], alm)]
        else:
            strucs = [
                class2tokens(struc, alm) for struc, alm in zip(structures, alignment)
            ]
        get_consensus(alignment, gaps=True)
        prostring = []
        for i in range(len(strucs[0])):
            row = [x[i] for x in strucs if x[i] != "-"]
            prostring += [row[0] if row else "+"]
        return [(i, p) for i, p in enumerate(prostring)]

    def reflexes_from_pos(
        self, position, taxa, current_taxa, alignment, missing, irregular
    ):
        reflexes = []
        for t in taxa:
            if t not in current_taxa:
                reflexes += [missing]
            else:
                reflex = alignment[current_taxa.index(t)][position]
                if "/" in reflex:
                    reflex = reflex.split("/")[1]
                elif reflex[0] in irregular:
                    reflex = missing
                reflexes += [reflex]
        return reflexes

    def _check(self):
        """
        Check for problematic patterns in the data.
        """
        errors = []
        for idx, struc, alm in self.iter_rows(self._structure, self._alignment):
            self[idx, self._structure] = self._str_type(struc)
            self[idx, self._alignment] = self._str_type(alm)
            if not len(self[idx, self._structure]) == len(
                [x for x in self[idx, self._alignment] if x != "-"]
            ):
                print(
                    idx,
                    self[idx, self._structure],
                    "|",
                    self[idx, self._alignment],
                    "|",
                    self[idx, "tokens"],
                )
                log.warning("alignment and structure do not match in {0}".format(idx))
                errors += [idx]
        return errors

    def get_sites(self):
        """
        Retrieve the alignment sites of interest for initial analysis.
        """
        sites, all_sites, taxa = (
            collections.OrderedDict(),
            collections.OrderedDict(),
            self.cols,
        )
        errors = self._check()
        if errors:
            raise ValueError("found {0} problems in the data".format(len(errors)))

        # iterate over all sites in the alignment
        visited = []
        for cogid, msa in pb(
            sorted(self.msa[self.ref].items()),
            desc="CoPaR: get_patterns()",
            total=len(self.msa[self.ref]),
        ):
            # get essential data: taxa, alignment, etc.
            _taxa = [t for t in taxa if t in msa["taxa"]]
            _idxs = {t: msa["taxa"].index(t) for t in _taxa}
            _alms = [msa["alignment"][_idxs[t]] for t in _taxa]
            _wlid = [msa["ID"][_idxs[t]] for t in _taxa]

            # store visited entries
            visited += msa["ID"]
            if len(_taxa) >= self.minrefs:
                if self._mode == "fuzzy":
                    _strucs = []
                    for _widx in _wlid:
                        _these_strucs = self[_widx, self._structure]
                        _strucs += [_these_strucs]
                else:
                    _strucs = [self[idx, self._structure] for idx in _wlid]
                positions = self.positions_from_prostrings(cogid, _wlid, _alms, _strucs)
                for pidx, pos in positions:
                    reflexes = self.reflexes_from_pos(
                        pidx, taxa, _taxa, _alms, self.missing, self.irregular
                    )
                    sites[cogid, pidx] = [pos, tuple(reflexes)]
            for pidx in range(len(_alms[0])):
                reflexes = self.reflexes_from_pos(
                    pidx, taxa, _taxa, _alms, self.missing, self.irregular
                )
                all_sites[cogid, pidx] = reflexes

        # add non-visited segments
        for idx in [i for i in self if i not in visited]:
            if self._mode == "fuzzy":
                for tt, ss, cogid in zip(
                    self[idx, self._segments].n,
                    self[idx, self._structure].n,
                    self[idx, self._ref],
                ):
                    for i, (t, s) in enumerate(zip(tt, ss)):
                        all_sites[cogid, i] = [
                            self.missing if tax != self[idx][self._colIdx] else t
                            for tax in self.cols
                        ]
            else:
                for i, (t, s) in enumerate(
                    zip(self[idx, self._segments], self[idx, self._structure])
                ):
                    all_sites[self[idx, self.ref], i] = [
                        self.missing if tax != self[idx][self._colIdx] else t
                        for tax in self.cols
                    ]

        self.sites = sites
        self.all_sites = all_sites

    def cluster_sites(self, match_threshold=1, score_mode="pairs"):
        """Cluster alignment sites using greedy clique cover.
        :param match_threshold: The threshold of matches for accepting two
            compatible columns.
        :param score_mode: select between "pairs", "coverage"

        .. note:: This algorithm follows the spirit of the Welsh-Powell algorithm for
           graph coloring. Since graph coloring is the inverse of clique
           partitioning, we can use the algorithm in the same spirit.

        """
        if not hasattr(self, "clusters"):
            self.clusters = collections.defaultdict(list)
            for (cogid, idx), (pos, ptn) in self.sites.items():
                self.clusters[pos, ptn] += [(cogid, idx)]
        clusters = self.clusters
        while True:
            prog = 0
            with pb(
                desc="CoPaR: cluster_sites()", total=len(self.clusters)
            ) as progress:
                sorted_clusters = sorted(
                    clusters.items(),
                    key=lambda x: (
                        score_patterns(
                            [self.sites[y][1] for y in x[1]], mode=score_mode
                        ),
                        len(x[1]),
                    ),
                    reverse=True,
                )
                out = []
                while sorted_clusters:
                    ((this_pos, this_cluster), these_vals), remaining_clusters = (
                        sorted_clusters[0],
                        sorted_clusters[1:],
                    )
                    queue = []
                    for (next_pos, next_cluster), next_vals in remaining_clusters:
                        match, mism = compatible_columns(
                            this_cluster,
                            next_cluster,
                            missing=self.missing,
                            gap=self.gap,
                        )
                        if (
                            this_pos == next_pos
                            and match >= match_threshold  # noqa: W503
                            and mism == 0  # noqa: W503
                        ):
                            this_cluster = consensus_pattern(
                                [this_cluster, next_cluster]
                            )
                            these_vals += next_vals
                        else:
                            queue += [((next_pos, next_cluster), next_vals)]
                    sorted_clusters = queue
                    out += [((this_pos, this_cluster), these_vals)]
                    progress.update(len(self.sites) - len(queue) - prog)
                    prog = len(self.sites) - len(queue)
                clusters = {tuple(a): b for a, b in out}
                alls = [c for c in clusters]
                match = 0
                for i, (_a, a) in enumerate(alls):
                    for j, (_b, b) in enumerate(alls):
                        if i < j and _a == _b:
                            ma, mi = compatible_columns(
                                a, b, missing=self.missing, gap=self.gap
                            )
                            if ma and not mi:
                                match += 1
                if not match:
                    break
                else:
                    log.warning(
                        "iterating, since {0} clusters can further be merged".format(
                            match
                        )
                    )
        self.clusters = clusters
        self.ordered_clusters = sorted(clusters, key=lambda x: len(x[1]))

    def sites_to_pattern(self, threshold=1):
        """Algorithm assigns alignment sites to patterns.

        Notes
        -----
        We rank according to general compatibility.
        """
        asites = collections.defaultdict(list)
        for consensus in pb(
            self.clusters, desc="CoPaR: sites_to_pattern()", total=len(self.clusters)
        ):
            sites = self.clusters[consensus]
            for cog, pos in sites:
                struc, pattern = self.sites[cog, pos]
                for strucB, consensusB in self.clusters:
                    ma, mi = compatible_columns(pattern, consensusB)
                    if struc == strucB and not mi and ma >= threshold:
                        asites[cog, pos] += [(ma, struc, consensusB)]
        self.patterns = asites

    def fuzziness(self):
        return sum([len(b) for a, b in self.patterns.items()]) / len(self.patterns)

    def irregular_patterns(self, accepted=2, matches=1, irregular_prefix="!"):
        """
        Try to assign irregular patterns to accepted patterns.

        Parameters
        ----------
        accepted : int (default=2)
            Minimal size of clusters that we regard as regular.

        """
        bad_clusters = [
            (clr, pts[0]) for clr, pts in self.clusters.items() if len(pts) == 1
        ]
        good_clusters = sorted(
            [(clr, pts) for clr, pts in self.clusters.items() if len(pts) >= accepted],
            key=lambda x: len(x[1]),
            reverse=True,
        )
        new_clusters = {clr: [] for clr, pts in good_clusters}
        irregular_patterns = []
        for clr, ptn in bad_clusters:
            if ptn.count(self.missing) <= 2:
                for clrB, pts in good_clusters:
                    match, mism = compatible_columns(clr[1], clrB[1])
                    if mism <= matches and match > matches:
                        new_clusters[clrB] += [clr]
                        irregular_patterns += [clr]
                        break
        # re-assign alignments to the data by adding the irregular character
        for key, value in sorted(
            new_clusters.items(), key=lambda x: len(x[1]), reverse=True
        ):
            if len(value) > 0:
                for i, pattern in enumerate(value):
                    pt = []
                    for lid, (a, b) in enumerate(zip(key[1], pattern[1])):
                        if a != b and self.missing not in [a, b]:
                            pt += [irregular_prefix + b]
                            # assign pattern to the corresponding alignments
                            for cogid, position in self.clusters[pattern]:
                                if self._mode == "fuzzy":
                                    word_indices = self.etd[self.ref][cogid][lid]
                                    if word_indices:
                                        for widx in word_indices:
                                            # get the position in the alignment
                                            alms = self[widx, self._alignment].n
                                            cog_pos = self[widx, self.ref].index(cogid)
                                            new_alm = alms[cog_pos]
                                            new_alm[position] = "{0}{1}/{2}".format(
                                                irregular_prefix, b, a
                                            )
                                            alms[cog_pos] = new_alm
                                            self[
                                                widx, self._alignment
                                            ] = self._str_type(
                                                " + ".join(
                                                    [" ".join(x) for x in alms]
                                                ).split()
                                            )
                                else:
                                    word_indices = self.etd[self.ref][cogid][lid]
                                    if word_indices:
                                        for widx in word_indices:
                                            alm = self._str_type(
                                                self[widx, self._alignment]
                                            )
                                            alm[position] = "{0}{1}/{2}".format(
                                                irregular_prefix, b, a
                                            )
                                            self[
                                                widx, self._alignment
                                            ] = self._str_type(" ".join(alm))
                        else:
                            pt += [b]

        self.ipatterns = new_clusters
        for pattern, data in [
            (a, b) for a, b in bad_clusters if a not in irregular_patterns
        ]:
            cogid, position = data
            if self._mode == "fuzzy":
                for indices in [idx for idx in self.etd[self.ref][cogid] if idx]:
                    for widx in indices:
                        cog_pos = self[widx, self.ref].index(cogid)
                        alms = self[widx, self._alignment].n
                        new_alm = alms[cog_pos]
                        new_alm[position] = "{0}{1}".format(
                            irregular_prefix, new_alm[position]
                        )
                        alms[cog_pos] = new_alm
                        self[widx, self._alignment] = self._str_type(
                            " + ".join([" ".join(x) for x in alms]).split()
                        )

        return new_clusters

    def load_patterns(self, patterns="patterns"):
        self.id2ptn = collections.OrderedDict()
        self.clusters = collections.OrderedDict()
        self.id2pos = collections.defaultdict(set)
        self.sites = collections.OrderedDict()
        # get the template
        template = [self.missing for m in self.cols]
        tidx = {self.cols[i]: i for i in range(self.width)}
        for idx, ptn, alm, struc, doc, cogs in self.iter_rows(
            patterns, self._alignment, self._structure, "doculect", self._ref
        ):
            if self._mode == "fuzzy":
                ptn = bt.lists(ptn)
                for i in range(len(alm.n)):
                    for j, (p, a) in enumerate(zip(ptn.n[i], alm.n[i])):
                        if not p == "0/n":
                            this_pattern = self.id2ptn.get(p, [t for t in template])
                            if this_pattern[tidx[doc]] == "Ø":
                                this_pattern[tidx[doc]] = a
                            self.id2ptn[p] = this_pattern
                            self.id2pos[p].add((cogs[i], j))
            else:
                for j, (p, a) in enumerate(zip(ptn, alm)):
                    if not p == "0/n":
                        this_pattern = self.id2ptn.get(p, [t for t in template])
                        if this_pattern[tidx[doc]] == "Ø":
                            this_pattern[tidx[doc]] = a
                        self.id2ptn[p] = this_pattern
                        self.id2pos[p].add((cogs, j))

        self.ptn2id = {tuple(v): k for k, v in self.id2ptn.items()}
        for k, v in self.id2ptn.items():
            self.clusters[tuple(v)] = list(self.id2pos[k])
            self.id2pos[k] = list(self.id2pos[k])
            for s in self.id2pos[k]:
                self.sites[s] = [(len(self.id2pos[k]), tuple(v))]

    def add_patterns(
        self, ref="patterns", irregular_patterns=False, proto=False, override=True
    ):
        """Assign patterns to a new column in the word list."""
        if not hasattr(self, "id2ptn"):
            self.id2ptn = {}
        if not hasattr(self, "pattern2id"):
            self.ptn2id = {}
        if proto:
            pidx = self.cols.index(proto)
        else:
            pidx = 0

        if irregular_patterns:
            new_clusters = collections.defaultdict(list)
            for reg, iregs in self.ipatterns.items():
                for cogid, position in self.clusters[reg]:
                    new_clusters[reg] += [(cogid, position)]
                for ireg in iregs:
                    for cogid, position in self.clusters[ireg]:
                        new_clusters[reg] += [(cogid, position)]
        else:
            new_clusters = self.clusters
        for pattern, rest in self.clusters.items():
            for cogid, position in rest:
                if (cogid, position) not in new_clusters[pattern]:
                    new_clusters[pattern] += [(cogid, position)]

        P = {
            idx: bt.lists(
                [
                    "0/n" if x not in rc("morpheme_separators") else "+"
                    for x in self[idx, self._alignment]
                ]
            )
            for idx in self
        }
        for i, ((struc, pattern), data) in enumerate(
            sorted(new_clusters.items(), key=lambda x: len(x), reverse=True)
        ):
            pattern_id = "{0}-{1}/{2}".format(
                i + 1, len(self.clusters[struc, pattern]), pattern[pidx]
            )
            self.id2ptn[pattern_id] = pattern
            self.ptn2id[pattern] = pattern_id
            for cogid, position in data:
                word_indices = [c for c in self.etd[self.ref][cogid] if c]
                for idxs in word_indices:
                    for idx in idxs:
                        if self._mode == "fuzzy":
                            pattern_position = self[idx, self.ref].index(cogid)
                            this_pattern = P[idx].n[pattern_position]
                            try:
                                this_pattern[position] = pattern_id
                                P[idx].change(pattern_position, this_pattern)
                            except:  # noqa: E722
                                log.warning("error in {0}".format(cogid))

                        else:
                            P[idx][position] = pattern_id
        self.add_entries(ref, P, lambda x: x, override=override)

    def write_patterns(self, filename, proto=False, irregular_patterns=False):
        if proto:
            pidx = self.cols.index(proto)
        else:
            pidx = 0

        if not hasattr(self, "id2ptn"):
            raise ValueError("You should run CoPaR.add_patterns first!")

        if irregular_patterns:
            new_clusters = collections.defaultdict(list)
            for (pos, reg), iregs in self.ipatterns.items():
                for cogid, position in self.clusters[pos, reg]:
                    new_clusters[pos, reg] += [(cogid, position)]
                for _, ireg in iregs:
                    ireg_ = list(ireg)
                    print(ireg_)
                    for i, (a, b) in enumerate(zip(reg, ireg)):
                        print(i, a, b)
                        if a != b and b != self.missing:
                            ireg_[i] = a + "/" + b
                    ireg_ = tuple(ireg_)
                    self.ptn2id[ireg_] = self.ptn2id[reg]
                    for cogid, position in self.clusters[pos, ireg]:
                        new_clusters[pos, ireg_] += [(cogid, position)]
        else:
            new_clusters = self.clusters
        for (struc, pattern), rest in self.clusters.items():
            for cogid, position in rest:
                if (cogid, position) not in new_clusters[struc, pattern]:
                    new_clusters[struc, pattern] += [(cogid, position)]
        text = "ID\tSTRUCTURE\tFREQUENCY\t{0}\t{1}\tCOGNATESETS\tCONCEPTS\n".format(
            self.cols[pidx], "\t".join([c for c in self.cols if c != self.cols[pidx]])
        )

        sound = ""
        idx = 0
        for (struc, pattern), entries in sorted(
            new_clusters.items(),
            key=lambda x: (x[0][0], x[0][1][pidx], len(x[1])),
            reverse=True,
        ):
            if sound != pattern[pidx]:
                sound = pattern[pidx]
                idx = 0
            concepts = []
            for x, y in entries:
                for entry in self.etd[self.ref][x]:
                    if entry:
                        for value in entry:
                            concepts += [self[value, "concept"]]
            concepts = " / ".join(sorted(set(concepts)))

            idx += 1
            text += "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\n".format(
                self.ptn2id[pattern].split("/")[0],
                struc,
                len(entries),
                pattern[pidx],
                "\t".join([p for i, p in enumerate(pattern) if i != pidx]),
                ", ".join(["{0}:{1}".format(x, y) for x, y in entries]),
                concepts,
            )
        pathlib.Path(filename).write_text(text, encoding="utf8")

    def purity(self):
        """
        Compute the purity of the cluster analysis.

        .. note:: The purity is here interpreted as the degree to which
           patterns are filled with non-missing values. In this sense, it
           indicates to which degree information is computed and to which
           degree information is already provided by the data itself.
        """

        def get_purity(patterns):
            all_sums = []
            for i in range(len(patterns[0])):
                col = [line[i] for line in patterns]
                subset = set(col)
                sums = []
                for itm in subset:
                    if itm != self.missing:
                        sums += [col.count(itm) ** 2]
                if sums:
                    sums = math.sqrt(sum(sums)) / len(col)
                else:
                    sums = 0
                all_sums += [sums]
            return sum(all_sums) / len(all_sums)

        graph = self.get_cluster_graph()
        purities = []
        for node, data in graph.nodes(data=True):
            patterns = []
            for neighbor in graph[node]:
                patterns += [graph.nodes[neighbor]["pattern"].split()]
            if patterns:
                purities += [get_purity(patterns)]
            else:
                purities += [0]
        return sum(purities) / len(purities)

    def get_cluster_graph(self):
        """
        Compute a graph of the clusters.

        .. note:: In the cluster graph, the sites in the alignments are the
           nodes and the edges are drawn between nodes assigned to the same
           pattern.
        """

        graph = nx.Graph()
        for (pos, ptn), sites in self.clusters.items():
            for site in sites:
                graph.add_node(
                    "{0[0]}-{0[1]}".format(site),
                    pattern=" ".join(ptn),
                    site=" ".join(self.sites[site][1]),
                )

        for ((s1, p1), ptn1), ((s2, p2), ptn2) in itertools.combinations(
            self.sites.items(), r=2
        ):
            if ptn1[0] == ptn2[0]:
                m, mm = compatible_columns(ptn1[1], ptn2[1])
                if m and not mm:
                    graph.add_edge("{0}-{1}".format(s1, p1), "{0}-{1}".format(s2, p2))
        return graph

    def upper_bound(self):
        """
        Compute upper bound for clique partitioning following Bhasker 1991.
        """
        degs = {s: 0 for s in self.sites}
        sings = {s: 0 for s in self.sites}
        for (nA, (posA, ptnA)), (nB, (posB, ptnB)) in itertools.combinations(
            self.sites.items(), r=2
        ):
            if posA == posB:
                m, n = compatible_columns(ptnA, ptnB)
                if n > 0:
                    degs[nA] += 1
                    degs[nB] += 1
                else:
                    sings[nA] += 1
                    sings[nB] += 1
            else:
                degs[nA] += 1
                degs[nB] += 1

        return max([b for a, b in degs.items() if sings[a] > 0])

    def predict_words(self, **kw):
        """
        Predict patterns for those cognate sets where we have missing data.

        .. note::

           Purity (one of the return values) measures how well a given sound
           for a given site is reflected by one single sound (rather than
           multiple patterns pointing to different sounds) for a given
           doculect. It may be seen as a control case for the purity of a given
           prediction: if there are many alternative possibilities, this means
           that there is more uncertainty regarding the reconstructions or
           predictions.

        """
        if not hasattr(self, "sites"):
            raise ValueError("You need to compute alignment sites first")

        minrefs = self.minrefs
        missing = self.missing
        samples = kw.get("samples", 3)

        # pre-analyse the data to get for each site the best patterns in ranked
        # form
        ranked_sites = {}
        ranked_clusters = sorted(
            [(s, p, len(f)) for (s, p), f in self.clusters.items()],
            key=lambda x: x[2],
            reverse=True,
        )
        for (cogid, pos), ptns in self.patterns.items():
            struc, ptn = self.sites[cogid, pos]
            missings = [i for i in range(self.width) if ptn[i] == missing]
            if (struc, ptn) in self.clusters:
                ranked_sites[cogid, pos] = [
                    (len(self.clusters[struc, ptn]), struc, ptn)
                ]
            else:
                ranked_sites[cogid, pos] = [(1, struc, ptn)]
            for strucB, ptnB, freq in ranked_clusters:
                m, mm = compatible_columns(ptn, ptnB)
                if struc == strucB and m >= 1 and mm == 0:
                    if len(missings) > len(
                        [ptnB[i] for i in missings if ptnB[i] == missing]
                    ):
                        ranked_sites[cogid, pos] += [(freq, strucB, ptnB)]

        purity = {site: {} for site in ranked_sites}

        preds = {}
        for cogid, msa in self.msa[self._ref].items():
            missings = [t for t in self.cols if t not in msa["taxa"]]
            if len(set(msa["taxa"])) >= minrefs:
                words = [bt.strings("") for m in missings]
                for i, m in enumerate(missings):
                    tidx = self.cols.index(m)
                    for j in range(len(msa["alignment"][0])):
                        segments = collections.defaultdict(int)
                        sidx = 0
                        if (cogid, j) in ranked_sites:
                            while True:
                                this_segment = ranked_sites[cogid, j][sidx][2][tidx]
                                score = ranked_sites[cogid, j][sidx][0]
                                if this_segment != missing:
                                    segments[this_segment] += score
                                sidx += 1
                                if sidx == len(ranked_sites[cogid, j]):
                                    break

                        if not (cogid, j) in purity:
                            purity[cogid, j] = {}

                        if not segments:
                            words[i] += ["Ø"]
                            purity[cogid, j][m] = 0
                        else:
                            purity[cogid, j][m] = math.sqrt(
                                sum(
                                    [
                                        (s / sum(segments.values())) ** 2
                                        for s in segments.values()
                                    ]
                                )
                            )
                            words[i] += [
                                "|".join(
                                    [
                                        s
                                        for s in sorted(
                                            segments,
                                            key=lambda x: segments[x],
                                            reverse=True,
                                        )
                                    ][:samples]
                                )
                            ]
                if words:
                    preds[cogid] = dict(zip(missings, words))

        pudity = {doc: [] for doc in self.cols}
        for site, docs in purity.items():
            for doc in docs:
                pudity[doc] += [purity[site][doc]]
        for doc, purs in pudity.items():
            if purs:
                pudity[doc] = sum(purs) / len(purs)
            else:
                pudity[doc] = 0

        return preds, purity, pudity



"""
Module provides methods for linguistic reconstruction.
"""
import itertools
import collections

from lingpy.align.sca import Alignments, get_consensus
from lingpy.sequence.sound_classes import prosodic_string, class2tokens
from lingpy.align.multiple import Multiple
from lingpy.align.pairwise import edit_dist, nw_align
from lingpy.evaluate.acd import _get_bcubed_score as get_bcubed_score
from lingpy.align.sca import normalize_alignment
import networkx as nx
from networkx.algorithms.clique import find_cliques
from lingpy import log

from lingrex.util import clean_sound, ungap, alm2tok


class CorPaRClassifier(object):

    def __init__(self, minrefs=2, missing=0, threshold=1):
        self.G = nx.Graph()
        self.missing = 0
        self.threshold = threshold

    def compatible(self, ptA, ptB):
        """
        Check for compatibility of two patterns.
        """
        res = {True: 0, False: 0}
        for a, b in zip(ptA, ptB):
            if a and b:
                res[a == b] += 1
        return res[True], res[False]

    def consensus(self, nodes):
        """
        Create a consensus pattern of multiple alignment sites.
        """
        cons = []
        for i in range(len(nodes[0])):
            nocons = True
            for node in nodes:
                if node[i] != self.missing:
                    cons += [node[i]]
                    nocons = False
                    break
            if nocons:
                cons += [self.missing]
        return tuple(cons)

    def fit(self, X, y):
        """
        Train the prediction of data in y with data in X.

        :param X: Two-dimensional array with observations.
        :param y: One-dimensional array with results.
        """
        # get identical patterns
        P = collections.defaultdict(list)
        for i, row in enumerate(X):
            P[tuple(row + [y[i]])] += [i]
        # make graph
        for (pA, vA), (pB, vB) in itertools.combinations(P.items(), r=2):
            match_, mismatch = self.compatible(pA, pB)
            if not mismatch and match_ >= self.threshold:
                if pA not in self.G:
                    self.G.add_node(pA, freq=len(vA))
                if pB not in self.G:
                    self.G.add_node(pB, freq=len(vB))
                self.G.add_edge(pA, pB, weight=match_)
        self.patterns = collections.defaultdict(collections.Counter)
        self.lookup = collections.defaultdict(collections.Counter)
        # get cliques
        for nodes in find_cliques(self.G):
            cons = self.consensus(list(nodes))
            self.patterns[cons[:-1]][cons[-1]] = len(nodes)
            for node in nodes:
                self.lookup[node[:-1]][cons[:-1]] += len(nodes)
        self.predictions = {
            ptn: counts.most_common(1)[0][0] for ptn, counts in self.patterns.items()}
        for ptn, counts in self.lookup.items():
            self.predictions[ptn] = self.predictions[counts.most_common(1)[0][0]]

        # make index of data points for quick search based on attested data
        self.ptnlkp = collections.defaultdict(list)
        for ptn in self.patterns:
            for i in range(len(ptn)):
                if ptn[i] != self.missing:
                    self.ptnlkp[i, ptn[i]] += [ptn]

    def predict(self, matrix):
        out = []
        for row in matrix:
            ptn = tuple(row)
            if ptn in self.predictions:
                out.append(self.predictions[ptn])
            else:
                candidates = collections.Counter()
                for i in range(len(ptn) - 1):
                    if ptn[i] != self.missing:
                        for ptnB in self.ptnlkp[i, ptn[i]]:
                            if ptnB not in candidates:
                                match_, mismatch = self.compatible(ptn, ptnB)
                                if match_ and not mismatch:
                                    candidates[ptnB] = match_ + len(ptn)
                                elif match_ - mismatch:
                                    candidates[ptnB] = match_ - mismatch
                if candidates:
                    self.predictions[tuple(row)] = self.predictions[candidates.most_common(1)[0][0]]
                    out += [self.predictions[tuple(row)]]
                else:
                    out += [self.missing]
        return out


class ReconstructionBase(Alignments):
    """
    Basic class for the phonological reconstruction.
    """
    def __init__(
            self, infile, target=None, ref="cogids", fuzzy=True,
            transcription="form", missing="Ø", gap="-"):
        Alignments.__init__(self, infile, fuzzy=fuzzy, ref=ref, transcription=transcription)
        self.target = target
        self.missing = missing
        self.gap = gap
        self.languages = [t for t in self.cols if t != target]
        self.target = target
        self.tgtidx = self.cols.index(target)
        self.lngidx = {t: self.cols.index(t) for t in self.languages}

    def iter_sequences(self, aligned=False):
        """
        Iterate over aligned or unaligned sequences with or without the target \
                sequence.
        """
        seq_ref = self._alignments if aligned else self._segments
        for cogid, idxs in self.etd[self._ref].items():
            if idxs[self.tgtidx]:
                if self._mode == "fuzzy":
                    target = self[idxs[self.tgtidx][0], seq_ref].n[
                        self[idxs[self.tgtidx][0], self._ref].index(cogid)]
                else:
                    target = self[idxs[self.tgtidx][0], seq_ref]
                alignment, languages = [], []
                for j, lng in enumerate(self.languages):
                    lidx = self.lngidx[lng]
                    if idxs[lidx]:
                        languages += [lng]
                        idx = idxs[lidx][0]
                        if self._mode == "fuzzy":
                            alm = self[idx, seq_ref].n[self[idx, self._ref].index(cogid)]
                        else:
                            alm = self[idx, seq_ref]
                        alignment.append([clean_sound(x) for x in alm])
                alignment.append([clean_sound(x) for x in target])
                if aligned:
                    alignment = normalize_alignment(alignment)
                languages.append(self.target)
                yield cogid, alignment, languages


class OneHot(object):
    """
    Create a one-hot-encoder from a matrix.
    """

    def __init__(self, matrix):
        self.vals = []
        for i in range(len(matrix[0])):
            cols = [row[i] for row in matrix]
            self.vals += [sorted(set(cols)) + ["?"]]

    def __call__(self, matrix):
        out = [[] for row in matrix]
        for i, vals in enumerate(self.vals):
            for j in range(len(matrix)):
                template = [0 for k in vals]
                try:
                    template[matrix[j][i]] = 1
                except IndexError:
                    template[-1] = 1
                out[j] += template
        return out


def transform_alignment(seqs,
                        languages,
                        all_languages,
                        align=True,
                        training=True,
                        missing="Ø",
                        gap="-",
                        startend=False,
                        prosody=False,
                        position=False,
                        firstlast=False):
    """
    Basic alignment function used for phonological reconstruction.
    """
    if align:
        seqs = [[s for s in seq if s != gap] for seq in seqs]
        msa = Multiple([[s for s in seq if s != gap] for seq in seqs])
        msa.prog_align()
        alms = [alm for alm in msa.alm_matrix]
    else:
        seqs = [[s for s in seq if s != gap] for seq in seqs]
        alms = normalize_alignment([s for s in seqs])
    if training:
        alms = ungap(alms, languages, languages[-1])
        these_seqs = seqs[:-1]
    else:
        these_seqs = seqs
    matrix = [[missing for x in all_languages] for y in alms[0]]
    for i in range(len(alms[0])):
        for j, lng in enumerate(languages):
            lidx = all_languages.index(lng)
            snd = clean_sound(alms[j][i])
            matrix[i][lidx] = snd
    if position:
        for i in range(len(matrix)):
            matrix[i] += [i]
    if startend:
        matrix[0] += [0]
        for i in range(1, len(matrix) - 1):
            matrix[i] += [1]
        if len(matrix) > 1:
            matrix[-1] += [2]
    if prosody:
        for i, c in enumerate(
                get_consensus(
                    [class2tokens(prosodic_string(seqs[j], _output="CcV"), alms[j])
                     for j in range(len(these_seqs))],
                    gaps=True)):
            matrix[i] += [c]
    if firstlast:
        if training:
            all_seqs = len(all_languages) - 1
        else:
            all_seqs = len(all_languages)
        for i, row in enumerate(matrix):
            for j in range(all_seqs):
                matrix[i] += [matrix[0][j], matrix[-1][j]]

    # for debugging
    for row in matrix:
        assert len(row) == len(matrix[0])
    return matrix


class PatternReconstructor(ReconstructionBase):
    """
    Automatic reconstruction with correspondence patterns.
    """

    def fit(self, clf=None, onehot=False, func=None, aligned=False):
        """
        Fit a classifier to the data.

        :param clf: a classifier with a predict function.
        """
        self.patterns = collections.defaultdict(lambda: collections.defaultdict(list))
        self.occurrences = collections.defaultdict(list)
        self.func = func or transform_alignment

        for cogid, alignment, languages in self.iter_sequences():
            if len(alignment) >= 2:
                matrix = self.func(
                    alignment,
                    languages,
                    self.languages + [self.target],
                    training=True)
                for i, row in enumerate(matrix):
                    ptn = tuple(row[:len(self.languages)] + row[len(self.languages) + 1:])
                    self.patterns[ptn][row[len(self.languages)]] += [
                        (cogid, i)]
                    for j, lng in enumerate(self.languages):
                        if row[j] not in [self.missing]:
                            self.occurrences[lng, j, row[j]] += [(cogid, i)]
                    for j in range(len(self.languages) + 1, len(row)):
                        self.occurrences["feature-{0}".format(j - 1), j - 1, row[j]] += [(cogid, i)]

        self.snd2idx = {(i, self.missing): 0 for i in range(len(matrix[0]))}
        for i in range(len(matrix[0])):
            self.snd2idx[i, self.gap] = 1

        idxtracker = {i: 2 for i in range(len(matrix[0]))}
        for lng, lidx, sound in self.occurrences:
            last_idx = idxtracker[lidx]
            if (lidx, sound) not in self.snd2idx:
                self.snd2idx[lidx, sound] = last_idx
                idxtracker[lidx] += 1

        self.tgt2idx = {}
        idx = 1
        for pattern in self.patterns:
            for sound in self.patterns[pattern]:
                if sound not in self.tgt2idx:
                    self.tgt2idx[sound] = idx
                    idx += 1

        self.matrix = []
        self.solutions = []
        for pattern, sounds in self.patterns.items():
            for sound, vals in sounds.items():
                tidx = self.tgt2idx[sound]
                row = []
                for i in range(len(pattern)):
                    sidx = self.snd2idx[i, pattern[i]]
                    row += [sidx]
                for cogid, idx in vals:
                    self.matrix += [row]
                    self.solutions += [tidx]
        self.dim = len(self.matrix[0])
        if clf is not None:
            self.clf = clf
        else:
            self.clf = CorPaRClassifier()
        log.info("fitting classifier")
        if onehot:
            self.onehot = OneHot(self.matrix)
            self.clf.fit(self.onehot(self.matrix), self.solutions)
        else:
            self.clf.fit(self.matrix, self.solutions)
        self.idx2tgt = {v: k for k, v in self.tgt2idx.items()}
        log.info("fitted the classifier")

    def predict(
            self, alignment, languages, unknown="?", onehot=False,
            desegment=True):
        """
        Predict a word form from an alignment.

        :param desegment: Return the form without gaps and ungapped tokens.
        """
        matrix = self.func(alignment, languages, self.languages, training=False)
        for row in matrix:
            assert len(row) == self.dim
        new_matrix = [[0 for char in row] for row in matrix]
        for i, row in enumerate(matrix):
            for j, char in enumerate(row):
                new_matrix[i][j] = self.snd2idx.get((j, char), 0)
        if hasattr(self, "onehot"):
            new_matrix = self.onehot(new_matrix)
        out = [self.idx2tgt.get(idx, unknown) for idx in self.clf.predict(new_matrix)]
        return alm2tok(out) if desegment else out


def eval_by_dist(data, func=None, **kw):
    """
    Evaluate by measuring distances between sequences.

    :param data: List of tuples with prediction and attested sequence.
    :param func: Alignment function (defaults to edit distance)

    :note: Defaults to the unnormalized edit distance.
    """
    func = func or edit_dist
    scores = []
    for seqA, seqB in data:
        if not seqA:
            seqA = ["?"]
        if not seqB:
            seqB = ["?"]
        scores += [func(seqA, seqB, **kw)]
    return sum(scores) / len(scores)


def eval_by_bcubes(data, func=None, **kw):
    """
    Evaluate by measuring B-Cubed F-scores.

    :param data: List of tuples with prediction and attested sequence.
    :param func: Alignment function (defaults to Needleman-Wunsch)
    """
    numsA, numsB = {"": 0}, {"": 0}
    func = func or nw_align
    almsA, almsB = [], []
    for seqA, seqB in data:
        if not seqA:
            seqA = ["?"]
        if not seqB:
            seqB = ["?"]
        almA, almB, score = func(seqA, seqB, **kw)
        for a, b in zip(almA, almB):
            if a not in numsA:
                numsA[a] = max(numsA.values()) + 1
            if b not in numsB:
                numsB[b] = max(numsB.values()) + 1
            almsA += [numsA[a]]
            almsB += [numsB[b]]
    p, r = get_bcubed_score(almsA, almsB), get_bcubed_score(almsB, almsA)
    return 2 * (p * r) / (p + r)



"""
Functions for partial colexification manipulations.
"""
import collections


def find_bad_internal_alignments(alignments, ref="cogids"):
    """
    Helper function discards wrongly assigned cross-semantic cognates.

    .. note:: The function essentially iterates over the alignments and picks
        out those in which the same language has the same cognate ID, and if
        the alignment itself differs, it assigns it a new cognate ID. It
        presupposes that the data has not been analyzed in search for
        cross-semantic cognates.
    """
    newIDs = {}

    def get_all_indices(lst):
        idxs = collections.defaultdict(list)
        for i, l in enumerate(lst):
            idxs[l] += [i]
        return idxs

    new_cogid = max(alignments.msa[ref]) + 1
    for cogid, msa in alignments.msa[ref].items():
        idxs = [i for t, i in get_all_indices(msa["taxa"]).items() if len(i) > 1]
        for idx in idxs:
            tups = [tuple(msa["alignment"][x]) for x in idx]
            if len(set(tups)) > 1:
                bestt = sorted(tups, key=lambda x: tups.count(x), reverse=True)[0]
                for x in idx:
                    if tuple(msa["alignment"][x]) != bestt:
                        newIDs[msa["ID"][x]] = (cogid, new_cogid)
                        new_cogid += 1

    for idx, (cogid, new_cogid) in newIDs.items():
        this_idx = alignments[idx, ref].index(cogid)
        alignments[idx, ref][this_idx] = new_cogid


def expand_alignment(msa, taxa, missing="Ø"):
    """
    Expand an alignment by adding a symbol for missing taxa.
    """
    out = []
    for taxon in taxa:
        if taxon in msa["taxa"]:
            tidx = msa["taxa"].index(taxon)
            out.append(
                [x.split("/")[1] if "/" in x else x for x in msa["alignment"][tidx]]
            )
        else:
            out.append(len(msa["alignment"][0]) * [missing])
    return out


def compatible(msa1, msa2, missing="Ø", gap="-"):
    """Compare two alignments and check whether they colexify."""
    matches = 0
    for line1, line2 in zip(msa1, msa2):
        if [x for x in line1 if x != gap] == [
            x for x in line2 if x != gap
        ] and missing not in line1 + line2:
            matches += 1
        else:
            if list(set(line1))[0] != missing and list(set(line2))[0] != missing:
                return False
    return matches


def merge_alignments(almA, almB, missing="Ø", gap="-"):
    """
    Merge two alignments which are compatible.
    """
    out = []
    missing_taxa = []
    for k, (a, b) in enumerate(zip(almA, almB)):
        if (
            len(set(a)) == 1
            and list(set(a))[0] == missing  # noqa: W503
            and len(set(b)) == 1  # noqa: W503
            and list(set(b))[0] == missing  # noqa: W503
        ):
            missing_taxa += [k]
    i, j = 0, 0
    while i < len(almA[0]) and j < len(almB[0]):
        colA, colB = [row[i] for row in almA], [row[j] for row in almB]
        if colA == colB:
            out += [colA]
            i += 1
            j += 1
        else:
            col = []
            for a, b in zip(colA, colB):
                if a == gap and a != b and b != missing:
                    ncol = []
                    for k, c in enumerate(colA):
                        if c == missing and k not in missing_taxa:
                            ncol += [gap]
                        else:
                            ncol += [c]
                    out += [ncol]
                    i += 1
                    col = []
                    break
                if b == gap and a != b and a != missing:
                    ncol = []
                    for k, c in enumerate(colB):
                        if c == missing and k not in missing_taxa:
                            ncol += [gap]
                        else:
                            ncol += [c]
                    out += [ncol]
                    j += 1
                    col = []
                    break

                col.append(b if a == missing else a)
            if col:
                out += [col]
                i += 1
                j += 1
    if i < len(almA[0]):
        ncol = []
        for k, c in enumerate([row[i] for row in almA]):
            if c == missing and k not in missing_taxa:
                ncol += [gap]
            else:
                ncol += [c]
        out += [ncol]
    elif j < len(almB[0]):
        ncol = []
        for k, c in enumerate([row[j] for row in almB]):
            if c == missing and k not in missing_taxa:
                ncol += [gap]
            else:
                ncol += [c]
        out += [ncol]

    nalm = []
    for i in range(len(out[0])):
        nalm += [[row[i] for row in out]]
    return nalm


def find_colexified_alignments(
    alignments, cognates="cogids", missing="Ø", ref="crossids"
):
    """
    Identify identical alignments in a dataset and label them as homophones.
    """

    queue = []
    for cogid, msa in sorted(
        alignments.msa[cognates].items(),
        key=lambda x: len(set(x[1]["taxa"])),
        reverse=True,
    ):
        queue += [(cogid, expand_alignment(msa, alignments.taxa, missing=missing))]

    merged = {}

    while queue:
        this_cogid, this_msa = queue.pop(0)
        deletes = []
        merged[this_cogid] = this_cogid
        for i, (other_cogid, other_msa) in enumerate(queue):
            if compatible(this_msa, other_msa) >= 1:
                this_msa = merge_alignments(this_msa, other_msa)
                merged[other_cogid] = this_cogid
                deletes += [i]

        for i in deletes[::-1]:
            del queue[i]

    # assemble the clusters now
    if alignments._mode == "fuzzy":
        alignments.add_entries(ref, cognates, lambda x: [merged.get(y, y) for y in x])
    else:
        alignments.add_entries(ref, cognates, lambda x: merged.get(x, x))



"""
Basic code for borrowing detection.
"""
import itertools
import collections

from lingpy import Pairwise
from lingpy.compare.partial import Partial
from lingpy.compare.lexstat import LexStat

import networkx as nx

from lingpy.util import pb


def internal_cognates(
    wordlist,
    family="family",
    partial=True,
    method="lexstat",
    runs=10000,
    threshold=0.50,
    smooth=1,
    ratio=(2, 1),
    vscale=0.5,
    restricted_chars="_",
    modes=[("global", -1, 0.5), ("overlap", -1, 0.5)],
    ref="autocogids",
    cluster_method="upgma",
    model="sca",
):
    """
    Cluster the data into cognate sets, but only inside each family.

    :param family: name of the column in which language family information can
       be found (defaults="family")
    """
    families = {wordlist[k, family] for k in wordlist}

    # split data into parts
    D = {k: {} for k in sorted(families)}
    for idx, fam in wordlist.iter_rows(family):
        D[fam][idx] = [cell for cell in wordlist[idx]]

    gcogid = 0
    G = {}
    for fam, data in D.items():
        data[0] = [h for h in wordlist.columns]
        if partial:
            lex = Partial(data, model=model)
            if method == "lexstat":
                lex.get_partial_scorer(
                    runs=runs,
                    smooth=smooth,
                    ratio=ratio,
                    vscale=vscale,
                    restricted_chars=restricted_chars,
                    modes=modes,
                )
            lex.partial_cluster(
                ref=ref,
                method=method,
                cluster_method=cluster_method,
                threshold=threshold,
            )
        else:
            lex = LexStat(data, model=model)
            if method == "lexstat":
                lex.get_scorer(
                    runs=runs,
                    smooth=smooth,
                    ratio=ratio,
                    vscale=vscale,
                    restricted_chars=restricted_chars,
                    modes=modes,
                )
            lex.cluster(
                ref=ref,
                method=method,
                cluster_method=cluster_method,
                threshold=threshold,
            )

        # prepare global cognate indicies
        if partial:
            C = {idx: len(lex[idx, ref]) * [0] for idx in lex}
            etd = lex.get_etymdict(ref=ref)
            for cogid, idxs in etd.items():
                for idx_ in idxs:
                    if idx_:
                        for idx in idx_:
                            cogids = lex[idx, ref]
                            C[idx][cogids.index(cogid)] = cogid + gcogid
        else:
            C = {idx: 0 for idx in lex}
            etd = lex.get_etymdict(ref=ref)
            for cogid, idxs in etd.items():
                for idx_ in idxs:
                    if idx_:
                        for idx in idx_:
                            C[idx] = cogid + gcogid
        for idx in lex:
            G[idx] = C[idx]
        gcogid += max(etd) + 1

    renumber = {}
    cogid = 1
    if partial:
        for idx, vals in G.items():
            f = wordlist[idx, family]
            new_cogids = []
            for v in vals:
                if (f, v) in renumber:
                    new_cogids += [renumber[f, v]]
                else:
                    renumber[f, v] = cogid
                    new_cogids += [cogid]
                    cogid += 1
            G[idx] = new_cogids
    else:
        for idx, val in G.items():
            f = wordlist[idx, family]
            if (f, val) not in renumber:
                renumber[f, val] = cogid
                cogid += 1
            G[idx] = renumber[f, val]

    wordlist.add_entries(ref, G, lambda x: x)


def external_cognates(
    wordlist,
    cognates="autocogid",
    ref="autoborid",
    threshold=0.3,
    segments="tokens",
    gop=-1,
    family="family",
    doculect="doculect",
    concept="concept",
    align_mode="overlap",
):
    """
    Compute language-external cognates and assign them to cognate sets.

    :param cognates: The column which holds previously calculated cognates.
    :param ref: The column which will store the new borrowing identifiers.
    :param family: The column storing family information.
    :param doculect: The column storing doculect information.
    """

    B = {}
    borid = 1
    # iterate over the concepts
    for concept in pb(wordlist.rows):
        idxs = wordlist.get_list(row=concept, flat=True)
        for idx in idxs:
            B[idx] = 0
        taxa = [wordlist[idx, doculect] for idx in idxs]
        famis = [wordlist[idx, family] for idx in idxs]
        if len(set(famis)) > 1:
            G = nx.Graph()
            tokens = [wordlist[idx, segments] for idx in idxs]
            cogids = [wordlist[idx, cognates] for idx in idxs]

            # assemble cogids to groups
            groups = collections.defaultdict(list)
            for i, d, t, c in zip(idxs, taxa, tokens, cogids):
                groups[c] += [(i, d, t)]

            for group, items in groups.items():
                G.add_node(
                    str(group),
                    concept=concept,
                    taxa=", ".join([t[1] for t in items]),
                    idxs=", ".join([str(t[0]) for t in items]),
                    family=wordlist[[t[0] for t in items][0], family],
                )

            # compare groups
            for (gA, iA), (gB, iB) in itertools.combinations(list(groups.items()), r=2):
                if G.nodes[str(gA)]["family"] != G.nodes[str(gB)]["family"]:
                    wpairs = [(a[2], b[2]) for a, b in itertools.product(iA, iB)]

                    pairs = Pairwise(wpairs)
                    pairs.align(distance=True, gop=gop, mode=align_mode)
                    dst = []
                    for i, p in enumerate(pairs._alignments):
                        dst += [p[2]]

                    dst = sum(dst) / len(dst)
                    if dst <= threshold:
                        G.add_edge(str(gA), str(gB), distance=dst)

            # components
            for i, comp in enumerate(nx.connected_components(G)):
                if len(comp) > 1:
                    table = []
                    for cogid in comp:
                        idxs = [int(x) for x in G.nodes[cogid]["idxs"].split(", ")]
                        for idx in idxs:
                            table += [
                                [
                                    wordlist[idx, doculect],
                                    wordlist[idx, concept],
                                    str(wordlist[idx, segments]),
                                    wordlist[idx, family],
                                    cogid,
                                ]
                            ]
                        for idx in idxs:
                            B[idx] = borid
                    borid += 1
    wordlist.add_entries(ref, B, lambda x: x)



"""
Utility functions for the lingrex package.
"""
import math
import pathlib

from lingpy import tokens2class, prosodic_string
from lingpy.align.sca import get_consensus
from lingpy import basictypes as bt
from lingpy.sequence.ngrams import get_n_ngrams


def lingrex_path(*comps):
    return str(pathlib.Path(__file__).parent.joinpath(*comps))


def bleu_score(word, reference, n=4, weights=None, trim=True):
    """
    Compute the BLEU score for predicted word and reference.

    :param word: the predicted word
    :param reference: the predicted reference
    :param n: the order of ngrams
    :param weights: list of weights, should be the same size as n
    :param trim: bool, decide to trim n-grams or not
    """

    if not weights:
        weights = [1 / n for x in range(n)]

    scores = []
    for i in range(1, n + 1):
        new_wrd = list(get_n_ngrams(word, i))
        new_ref = list(get_n_ngrams(reference, i))
        if trim and i > 1:
            new_wrd = new_wrd[i - 1:-(i - 1)]
            new_ref = new_ref[i - 1:-(i - 1)]

        clipped, divide = [], []
        for itm in set(new_wrd):
            clipped += [new_ref.count(itm)]
            divide += [new_wrd.count(itm)]
        scores += [sum(clipped) / sum(divide)]

    # calculate arithmetic mean
    out_score = 1
    for weight, score in zip(weights, scores):
        out_score = out_score * (score ** weight)

    bp = 1 if len(word) > len(reference) else math.e ** (1 - (len(reference) / len(word)))
    return bp * (out_score ** (1 / sum(weights)))


def clean_sound(sound):
    """
    Get rid of "a/b" notation for sound segments.
    """
    return ".".join([s.split('/')[1] if "/" in s else s for s in sound.split('.')])


def alm2tok(seq, gap="-"):
    """Turn an alignment into a sequence."""
    return [clean_sound(x) for x in unjoin(seq) if x != gap]


def unjoin(seq):
    """
    Turn segments joined by a dot into unjoined segments.
    """
    out = []
    for itm in seq:
        out += itm.split('.')
    return out


def ungap(alignment, languages, proto):
    """
    Trim an MSA to remove all gaps in the target sequence.

    :examples:
      >>> ungap([['a', 'b'], ['x', '-'], ['y', '-']], ['proto', 'l1', 'l2'], 'proto')
      ... [['a.b'], ['x'], ['y']]
      >>> ungap([['a', 'b'], ['x', '-'], ['y', 'h']], ['proto', 'l1', 'l2'], 'proto')
      ... [['a', 'b'], ['x', '-'], ['y', 'h']]
    """
    pidxs = [i for i, taxon in enumerate(languages) if taxon == proto]
    merges = []
    for i in range(len(alignment[0])):  # go through the rows of the alignment ...
        col = [row[i] for row in alignment]
        # ... looking for gap-only alignments (in non-proto languages):
        if {site for j, site in enumerate(col) if j not in pidxs} == {'-'}:
            merges += [i]
    if not merges:
        return alignment
    new_alms = []
    for i, row in enumerate(alignment):
        new_alm, mergeit, started = [], False, True
        for j, cell in enumerate(row):
            if j in merges or mergeit:
                mergeit = False
                if not started:  # j != 0:
                    if cell != "-":
                        new_alm[-1] += '.' + cell if new_alm[-1] else cell
                else:
                    mergeit = True
                    new_alm.append("" if cell == "-" else cell)
            else:
                started = False
                new_alm.append(cell)
        new_alms.append([cell or "-" for cell in new_alm])
    return new_alms


def add_structure(
        wordlist, model="cv", segments="tokens", structure="structure", ref="cogid", gap="-"):
    """Add structure to a wordlist to make sure correspondence patterns can be
    inferred"""
    if model not in ["cv", "c", "CcV", "ps", "nogap"]:
        raise ValueError("[i] you need to select a valid model")
    D = {}
    if model == "cv":
        for idx, tks in wordlist.iter_rows(segments):
            D[idx] = " ".join(tokens2class(tks, "cv")).lower()

    if model == "c":
        for idx, tks in wordlist.iter_rows(segments):
            D[idx] = (
                " ".join(tokens2class(tks, "cv"))
                .lower()
                .replace("v", "c")
                .replace("t", "c")
            )
    if model == "nogap":
        assert hasattr(wordlist, "msa")
        for cogid, msa in wordlist.msa[ref].items():
            cons = [
                "c" if c != gap else gap
                for c in get_consensus(msa["alignment"], gaps=True)
            ]
            for idx, alm in zip(msa["ID"], msa["alignment"]):
                struc = []
                for a, b in zip(cons, alm):
                    if b != "-":
                        struc += [a]
                D[idx] = " ".join(struc)
        for idx, tks in wordlist.iter_rows(segments):
            if idx not in D:
                D[idx] = " ".join(["c" if c != "+" else c for c in tks])
    if model == "CcV":
        for idx, tks in wordlist.iter_rows(segments):
            D[idx] = " ".join(
                list(prosodic_string(tks, _output="CcV").replace("_", "+"))
            )
    if model == "ps":
        for idx, tks in wordlist.iter_rows(segments):
            D[idx] = " ".join(list(prosodic_string(tks)))

    if hasattr(wordlist, "_mode") and wordlist._mode == "fuzzy":
        struc_ = bt.lists
    else:
        struc_ = bt.strings
    wordlist.add_entries(structure, D, lambda x: struc_(x))



"""
Various phonetic alignment functions.
"""
from lingpy import basictypes as bt


def gap_free_pairwise(seqA, seqB, syllables=None, gap="-"):
    """
    Carry out a gap-free alignment in which segments are merged instead of gapped.
    """
    syllables = [] if syllables is None else syllables
    start = True
    merge = False
    outA, outB = [], []
    for i, (charA, charB) in enumerate(zip(seqA, seqB)):
        if i in syllables:
            start = True
        if start and charB == gap:
            outA.append(charA + ">")
            merge = True
        elif not merge and charB == gap:
            outA[-1] += "<" + charA
        elif merge:
            if charB == gap:
                outA[-1] += charA + ">"
            else:
                outA[-1] += charA
                outB.append(charB)
                merge = False
        else:
            outA.append(charA)
            outB.append(charB)
        start = False
    return outA, outB


def align_to_template(sequence, structures, template, gap="-"):
    """
    Align a sequence to a template.
    """
    if (len(sequence) != len(structures)) or (len(template) < len(sequence)):
        raise ValueError(
            "sequence {0} and structure {1} have different length".format(
                repr(sequence), repr(structures)
            )
        )
    if len([x for x in structures if x not in template]) != 0:
        raise ValueError(
            "{0} items in the structure {1} is not in the template".format(
                len([x for x in structures if x not in template]), repr(structures)
            )
        )

    out = []
    idxA, idxB = 0, 0
    while idxB < len(template):
        if idxA < len(sequence):
            segment, structure = sequence[idxA], structures[idxA]
        else:
            segment, structure = gap, ""
        current_structure = template[idxB]
        if current_structure == structure:
            out.append(segment)
            idxA += 1
        else:
            out.append(gap)
        idxB += 1

    return out


def shrink_alignments(alignments):
    """
    Remove columns from alignment which all consist of gaps.
    """
    excludes = []
    for i in range(len(alignments[0])):
        col = set([line[i] for line in alignments])
        if "-" in col and len(col) == 1:
            excludes.append(i)
    return [
        [site for i, site in enumerate(alignment) if i not in excludes]
        for alignment in alignments
    ]


def shrink(tokens, structures, converter):
    """
    Shrink tokens according to the converter.

    .. note:: Works only for shrinking two structure elements so far.
    """
    outt, outs = [], []
    sm, merge = None, False
    for i in range(len(tokens)):
        if i > 0:
            sm = " ".join([structures[i - 1], structures[i]])
            if sm in converter:
                outt += [tokens[i - 1] + tokens[i]]
                outs += [converter[sm]]
                merge = True
            elif not merge:
                outt += [tokens[i - 1]]
                outs += [converter.get(structures[i - 1], structures[i - 1])]
            else:
                merge = False
    if sm not in converter:
        outt += [tokens[i]]
        outs += [converter.get(structures[i], structures[i])]
    return outt, outs


def shrink_template(
    wordlist,
    structure="structure",
    segments="tokens",
    converter={"i m": "I", "i": "I", "n c": "R", "n": "R", "c": "R"},
    new_structure="structure2",
    new_tokens="tokens2",
    override=False,
):
    """
    Reduce a template by merging certain parts of the structure.
    """
    D = {}
    for idx, strucs, tokens in wordlist.iter_rows(structure, segments):
        D[idx] = shrink(tokens, strucs, converter)
    wordlist.add_entries(new_structure, D, lambda x: bt.lists(x[1]), override=override)
    wordlist.add_entries(new_tokens, D, lambda x: bt.lists(x[0]), override=override)


def template_alignment(
    wordlist,
    ref="cogid",
    template="CCCCVVccccT_CCCCVVccccT_CCCCVVccccT_CCCCVVccccT_CCCCvvT",
    structure="structure",
    fuzzy=False,
    segments="tokens",
    gap="-",
    alignment="alignment",
    override=True,
):
    """
    Function aligns the cognate sets in a wordlist to a template.
    """

    for idx, tokens, structures in wordlist.iter_rows(segments, structure):
        wordlist[idx, segments], wordlist[idx, structure] = bt.lists(tokens), bt.lists(
            structures
        )

    etd = wordlist.get_etymdict(ref)
    A = {}
    if not fuzzy:
        for cogid, vals in etd.items():
            idxs = []
            for val in vals:
                if val:
                    idxs += val
            alignments = shrink_alignments(
                [
                    align_to_template(
                        wordlist[idx, segments],
                        wordlist[idx, structure],
                        template,
                        gap=gap,
                    )
                    for idx in idxs
                ]
            )
            for idx, alm in zip(idxs, alignments):
                A[idx] = alm
    if fuzzy:
        cogid2alm = {}
        # only align the first item
        for cogid, vals in etd.items():
            idxs, alms, strucs = [], [], []
            for val in vals:
                if val:
                    idxs += val
                    alms += [
                        wordlist[idx, segments].n[wordlist[idx, ref].index(cogid)]
                        for idx in val
                    ]
                    strucs += [
                        wordlist[idx, structure].n[wordlist[idx, ref].index(cogid)]
                        for idx in val
                    ]
            alignments = shrink_alignments(
                [
                    align_to_template(alm, struc, template, gap=gap)
                    for alm, struc in zip(alms, strucs)
                ]
            )
            for idx, alm in zip(idxs, alignments):
                cogid2alm[cogid, idx] = " ".join(alm)
        # second iteration, add the alignments per cogid
        for idx, cogids in wordlist.iter_rows(ref):
            A[idx] = bt.lists(
                " + ".join([cogid2alm.get((cogid, idx), "?") for cogid in cogids])
            )
    wordlist.add_entries(alignment, A, lambda x: x, override=override)



"""
Operations with cognate sets.
"""
import collections

from clldutils.text import strip_brackets, split_text
import lingpy


def common_morpheme_cognates(
    wordlist, cognates="cogids", ref="autoid", morphemes="automorphemes", override=True
):
    """
    Convert partial cognates to full cognates.
    """

    C, M = {}, {}
    current = 1
    for concept in wordlist.rows:
        base = split_text(strip_brackets(concept))[0].upper().replace(" ", "_")
        idxs = wordlist.get_list(row=concept, flat=True)
        cogids = collections.defaultdict(list)
        for idx in idxs:
            M[idx] = [c for c in wordlist[idx, cognates]]
            for cogid in lingpy.basictypes.ints(wordlist[idx, cognates]):
                cogids[cogid] += [idx]
        for i, (cogid, idxs) in enumerate(
            sorted(cogids.items(), key=lambda x: len(x[1]), reverse=True)
        ):
            for idx in idxs:
                if idx not in C:
                    C[idx] = current
                    M[idx][M[idx].index(cogid)] = base
                else:
                    M[idx][M[idx].index(cogid)] = "_" + base.lower()
            current += 1
    wordlist.add_entries(ref, C, lambda x: x)
    if morphemes:
        wordlist.add_entries(morphemes, M, lambda x: x, override=override)


def salient_cognates(
    wordlist, cognates="cogids", ref="newcogid", morphemes="morphemes", override=True
):
    """
    Convert partial cognates to full cognates ignoring non-salient cognate sets.
    """

    lookup, D = {}, {}
    for idx, cogids, morphemes in wordlist.iter_rows(cognates, morphemes):
        selected_cogids = []
        for cogid, morpheme in zip(cogids, morphemes):
            if not morpheme.startswith("_"):
                selected_cogids += [cogid]
        salient = tuple(selected_cogids)
        if salient in lookup:
            D[idx] = lookup[salient]
        elif D.values():
            next_cogid = max(D.values()) + 1
            lookup[salient] = next_cogid
            D[idx] = next_cogid
        else:
            lookup[salient] = 1
            D[idx] = 1

    wordlist.add_entries(ref, D, lambda x: x, override=override)



from setuptools import setup, find_packages


setup(
        name='lingrex',
        description="Linguistic reconstruction with LingPy",
        long_description=open('README.md').read(),
        author='Johann-Mattis List and Robert Forkel',
        url='https://github.com/lingpy/lingrex',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
        ],
        version='1.2.0',
        packages=find_packages(where='src'),
        package_dir={'': 'src'},
        install_requires=['lingpy>=2.6.8'],
        extras_require={
            "test": ["pytest", "coverage", "pytest-mock", "pytest-cov"],
            "dev": ["wheel", "twine", "tox", "black", "flake8"],
        },
        keywords="historical linguistics, computational linguistics, computer-assisted language comparison"
        )
